module Permutation //begin architecture Permutation
  (
   input wire             clk,
   input wire             rst,
   input wire             start,
   input wire             a_rounds, //-- One: 12 rounds, Zero: 6 rounds
   input wire   [  63: 0]    Sr, //-- Rate
   input wire   [ 127: 0]   Sc0, Sc1, //-- Capacity
   output logic [  63: 0]  pSr, //-- Permuted rate
   output logic [ 127: 0] pSc0, pSc1, //-- Permuted capacity
   output logic           done
   );
   //-- Architecture
   //---------------------------------------------------------------------------------
   
   //-- Types ---------------------------------------------------------------------
   /*  type fsm is (*/
   typedef enum                   {s_idle, s_process, s_done} fsm_t;
   
   //-- Signals ------------------------------------------------------------------
   logic [  63:   0]      S0_o, S1_o, S2_o, S3_o, S4_o; //-- Permutation output after being registered
   logic [  63:   0]      x2_RC; //-- Third word after round constant addition
   logic [  63:   0]      x0_s, x1_s, x2_s, x3_s, x4_s; //-- State words after substitution
   logic [  63:   0]      x0_p, x1_p, x2_p, x3_p, x4_p; //-- State words after permuattion
   logic [   3:   0]      round, next_round; //-- Round number
   logic [3:0]                RC_i;            // TODO: specify a range in SV? natural range 0 to 11; */ //-- Round constant index
   fsm_t                  state, next_state; //-- State machine signals
   logic                  S_sel                           ; //-- Selector of mux
   logic [7:0]            RC_selected;
   
   //-- Constants ----------------------------------------------------------------
   /*    type RoundConstant is array (0 to 11) of std_logic_vector(7 downto 0); */ //-- Round constants
   /*    constant RC : RoundConstant := (x"f0", x"e1", x"d2", x"c3", x"b4", x"a5",*/
   /*                                    x"96", x"87", x"78", x"69", x"5a", x"4b");*/
   logic [7:0]            RC [0:15];

   assign RC[0] =  8'hf0;
   assign RC[1] =  8'he1;
   assign RC[2] =  8'hd2;
   assign RC[3] =  8'hc3;
   assign RC[4] =  8'hb4;
   assign RC[5] =  8'ha5;
   assign RC[6] =  8'h96;
   assign RC[7] =  8'h87;
   assign RC[8] =  8'h78;
   assign RC[9] =  8'h69;
   assign RC[10] = 8'h5a;
   assign RC[11] = 8'h4b;

   assign RC[12] = 8'h00;
   assign RC[13] = 8'h00;
   assign RC[14] = 8'h00;
   assign RC[15] = 8'h00;
   
   localparam              zero56 = 56'b0;
   
   //---------------------------------------------------------------------------------
   /*begin*/
   //-- Outputs assignments
   assign pSr = S0_o;
   assign pSc0 = {S1_o  ,  S2_o};
   assign pSc1 = {S3_o  ,  S4_o};

   assume property (@(posedge clk) RC_i<12);
   //assume property (@(posedge clk) $stable(Sr));
   //assume property (@(posedge clk) $stable(Sc0));
   //assume property (@(posedge clk) $stable(Sc1));
   
   //-- Addition of constants
   assign RC_i = (a_rounds  ==  1'b1) ? round : {round + 4'd6}; //TODO: check if any modulo arith based on VHDL being integer from 0 to 11, but I don't think round+6 is used for round>5 anyway
   assign RC_selected = RC[RC_i];
   assign x2_RC = S2_o ^ {zero56,RC_selected};
   
   //-- Substitution layer
   Sub_layer ps (// library work
     /*        port map(*/
     .x0( S0_o),
     .x1( S1_o),
     .x2( x2_RC),
     .x3( S3_o),
     .x4( S4_o),
     .Sx0( x0_s),
     .Sx1( x1_s),
     .Sx2( x2_s),
     .Sx3( x3_s),
     .Sx4( x4_s)
       );
   
   //-- Diffusion layer
   Diff_layer pl (// library work
     /*        port map(*/
     .x0( x0_s),
     .x1( x1_s),
     .x2( x2_s),
     .x3( x3_s),
     .x4( x4_s),
     .Dx0( x0_p),
     .Dx1( x1_p),
     .Dx2( x2_p),
     .Dx3( x3_p),
     .Dx4( x4_p)
       );
   
   //-- Clock Process
   always_ff @ (posedge clk) begin: perm_clk 
      
        //if (1'b1 /*rising_edge(clk)*/) begin /*        if rising_edge(clk) then */ //-- Registering signals
           if (rst  ==  1'b1) begin
              state <= s_idle;
           end else begin
              state <= next_state;
           end
           round <= next_round;
           if (S_sel  ==  1'b1) begin
              S0_o <= x0_p;
              S1_o <= x1_p;
              S2_o <= x2_p;
              S3_o <= x3_p;
              S4_o <= x4_p;
            end else begin
               S0_o <= Sr;
               S1_o <= Sc0[127:64];
               S2_o <= Sc0[63:0];
               S3_o <= Sc1[127:64];
               S4_o <= Sc1[63:0];
            end
        //end
   end //process perm_clk
   
   //-- State machine
   always_comb begin:  perm_fsm 
      
      //-- Default values
      next_state = s_idle;
      next_round = '0;
      S_sel = 1'b0;
      done = 1'b0;
      
      //-- Case statement
      case (state)
        s_idle  : begin //-- Wait for start
           if (start  ==  1'b1) begin
              next_state = s_process;
           end else begin
              next_state = s_idle;
           end
        end
        s_process  : begin //-- permutation
           S_sel = 1'b1;
           next_round = {round + 4'b1};
           if ((a_rounds  ==  1'b1 && round  ==  4'd11) || (a_rounds  ==  1'b0 && round  ==  4'd5)) begin //-- pa = 12, pb = 6
              next_round = '0;
              next_state = s_done;
           end else begin
              next_state = s_process;
           end
        end
        s_done  : begin //-- Done permutation
           done = 1'b1;
           if (start  ==  1'b1) begin
              next_state = s_process;
           end else begin
              next_state = s_idle;
           end
        end
      endcase
   end //process perm_fsm
   
endmodule //architecture Behavioral
